# Copyright (C) 2013, Red Hat, Inc.
# All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from pyasn1 import error
from pyasn1.codec.der import decoder, encoder
from pyasn1.type import char, namedtype, tag, univ

from kdcproxy.exceptions import ASN1ParsingError, ParsingError


class ProxyMessageKerberosMessage(univ.OctetString):
    tagSet = univ.OctetString.tagSet.tagExplicitly(
        tag.Tag(tag.tagClassContext, tag.tagFormatConstructed, 0)
    )


class ProxyMessageTargetDomain(char.GeneralString):
    tagSet = char.GeneralString.tagSet.tagExplicitly(
        tag.Tag(tag.tagClassContext, tag.tagFormatConstructed, 1)
    )


class ProxyMessageDCLocateHint(univ.Integer):
    tagSet = univ.Integer.tagSet.tagExplicitly(
        tag.Tag(tag.tagClassContext, tag.tagFormatSimple, 2)
    )


class ProxyMessage(univ.Sequence):
    pretty_name = 'KDC-PROXY-MESSAGE'

    componentType = namedtype.NamedTypes(
        namedtype.NamedType('message', ProxyMessageKerberosMessage()),
        namedtype.OptionalNamedType('realm', ProxyMessageTargetDomain()),
        namedtype.OptionalNamedType('flags', ProxyMessageDCLocateHint())
    )


class ASREQ(univ.Sequence):
    pretty_name = 'AS-REQ'

    tagSet = univ.Sequence.tagSet.tagExplicitly(
        tag.Tag(tag.tagClassApplication, tag.tagFormatSimple, 10)
    )


class TGSREQ(univ.Sequence):
    pretty_name = 'TGS-REQ'

    tagSet = univ.Sequence.tagSet.tagExplicitly(
        tag.Tag(tag.tagClassApplication, tag.tagFormatSimple, 12)
    )


class APREQ(univ.Sequence):
    pretty_name = 'AP-REQ'

    tagSet = univ.Sequence.tagSet.tagExplicitly(
        tag.Tag(tag.tagClassApplication, tag.tagFormatSimple, 14)
    )


class KRBPriv(univ.Sequence):
    pretty_name = 'KRBPRiv'

    tagSet = univ.Sequence.tagSet.tagExplicitly(
        tag.Tag(tag.tagClassApplication, tag.tagFormatSimple, 21)
    )


def decode_proxymessage(data):
    try:
        req, tail = decoder.decode(data, asn1Spec=ProxyMessage())
    except error.PyAsn1Error as e:
        raise ASN1ParsingError(e)
    if tail:
        raise ParsingError("Invalid request.")
    message = req.getComponentByName('message').asOctets()
    realm = req.getComponentByName('realm')
    if realm.hasValue():
        try:  # Python 3.x
            realm = str(realm, "utf-8")
        except TypeError:  # Python 2.x
            realm = str(realm)
    else:
        realm = None
    flags = req.getComponentByName('flags')
    flags = int(flags) if flags.hasValue() else None
    return message, realm, flags


def encode_proxymessage(data):
    rep = ProxyMessage()
    rep.setComponentByName('message', data)
    return encoder.encode(rep)


def try_decode(data, cls):
    try:
        req, tail = decoder.decode(data, asn1Spec=cls())
    except error.PyAsn1Error as e:
        raise ASN1ParsingError(e)
    if tail:
        raise ParsingError("%s request has %d extra bytes." %
                           (cls.pretty_name, len(tail)))
    return cls.pretty_name



# Copyright (C) 2013, Red Hat, Inc.
# All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import os
import struct

from kdcproxy.exceptions import ParsingError

ASN1MOD = os.environ.get('KDCPROXY_ASN1MOD')

if ASN1MOD is None:
    try:
        from asn1crypto.version import __version_info__ as asn1crypto_version
    except ImportError:
        asn1crypto_version = None
    else:
        if asn1crypto_version >= (0, 22, 0):
            ASN1MOD = 'asn1crypto'
    if ASN1MOD is None:
        try:
            __import__('pyasn1')
        except ImportError:
            pass
        else:
            ASN1MOD = 'pyasn1'

if ASN1MOD == 'asn1crypto':
    from kdcproxy import parse_asn1crypto as asn1mod
elif ASN1MOD == 'pyasn1':
    from kdcproxy import parse_pyasn1 as asn1mod
else:
    raise ValueError("Invalid KDCPROXY_ASN1MOD='{}'".format(ASN1MOD))


class ProxyRequest(object):
    TYPE = None
    OFFSET = 4

    @classmethod
    def parse(cls, data):
        request, realm, _ = asn1mod.decode_proxymessage(data)

        # Check the length of the whole request message.
        (length, ) = struct.unpack("!I", request[0:4])
        if length + 4 != len(request):
            raise ParsingError("Invalid request length.")

        for subcls in cls.__subclasses__():
            try:
                return subcls.parse_request(realm, request)
            except ParsingError:
                pass

        raise ParsingError("Invalid request.")

    @classmethod
    def parse_request(cls, realm, request):
        pretty_name = asn1mod.try_decode(request[cls.OFFSET:], cls.TYPE)
        return cls(realm, request, pretty_name)

    def __init__(self, realm, request, pretty_name):
        self.realm = realm
        self.request = request
        self.pretty_name = pretty_name

    def __str__(self):
        return "%s %s (%d bytes)" % (self.realm, self.pretty_name,
                                     len(self.request) - 4)


class TGSProxyRequest(ProxyRequest):
    TYPE = asn1mod.TGSREQ


class ASProxyRequest(ProxyRequest):
    TYPE = asn1mod.ASREQ


class KPASSWDProxyRequest(ProxyRequest):
    TYPE = asn1mod.APREQ
    OFFSET = 10

    @classmethod
    def parse_request(cls, realm, request):
        # Check the length count in the password change request, assuming it
        # actually is a password change request.  It should be the length of
        # the rest of the request, including itself.
        (length, ) = struct.unpack("!H", request[4:6])
        if length != len(request) - 4:
            raise ParsingError("Parsing the KPASSWD request length failed.")

        # Check the version number in the password change request, assuming it
        # actually is a password change request.  Officially we support version
        # 1, but 0xff80 is used for set-password, so try to accept that, too.
        (version, ) = struct.unpack("!H", request[6:8])
        if version != 0x0001 and version != 0xff80:
            raise ParsingError("The KPASSWD request is an incorrect version.")

        # Read the length of the AP-REQ part of the change request.  There
        # should be at least that may bytes following this length, since the
        # rest of the request is the KRB-PRIV message.
        (length, ) = struct.unpack("!H", request[8:10])
        if length > len(request) - 10:
            raise ParsingError("The KPASSWD request appears to be truncated.")

        # See if the tag looks like an AP request, which would look like the
        # start of a password change request. The rest of it should be a
        # KRB-PRIV message.
        asn1mod.try_decode(request[10:length + 10], asn1mod.APREQ)
        asn1mod.try_decode(request[length + 10:], asn1mod.KRBPriv)

        self = cls(realm, request, "KPASSWD-REQ")
        self.version = version
        return self

    def __str__(self):
        tmp = super(KPASSWDProxyRequest, self).__str__()
        tmp += " (version 0x%04x)" % self.version
        return tmp


def decode(data):
    return ProxyRequest.parse(data)


def encode(data):
    return asn1mod.encode_proxymessage(data)



# Copyright (C) 2017, Red Hat, Inc.
# All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from asn1crypto import core

from kdcproxy.exceptions import ASN1ParsingError


APPLICATION = 1


class KerberosString(core.GeneralString):
    """KerberosString ::= GeneralString (IA5String)

    For compatibility, implementations MAY choose to accept GeneralString
    values that contain characters other than those permitted by
    IA5String...
    """


class Realm(KerberosString):
    """Realm ::= KerberosString
    """


class ProxyMessage(core.Sequence):
    pretty_name = 'KDC-PROXY-MESSAGE'

    _fields = [
        ('kerb-message', core.OctetString, {
            'explicit': 0}),
        ('target-domain', Realm, {
            'explicit': 1, 'optional': True}),
        ('dclocator-hint', core.Integer, {
            'explicit': 2, 'optional': True}),
    ]


class ASREQ(core.Sequence):
    pretty_name = 'AS-REQ'

    explicit = (APPLICATION, 10)


class TGSREQ(core.Sequence):
    pretty_name = 'TGS-REQ'

    explicit = (APPLICATION, 12)


class APREQ(core.Sequence):
    pretty_name = 'AP-REQ'

    explicit = (APPLICATION, 14)


class KRBPriv(core.Sequence):
    pretty_name = 'KRBPRiv'

    explicit = (APPLICATION, 21)


def decode_proxymessage(data):
    req = ProxyMessage.load(data, strict=True)
    message = req['kerb-message'].native
    realm = req['target-domain'].native
    try:  # Python 3.x
        realm = str(realm, "utf-8")
    except TypeError:  # Python 2.x
        realm = str(realm)
    flags = req['dclocator-hint'].native
    return message, realm, flags


def encode_proxymessage(data):
    rep = ProxyMessage()
    rep['kerb-message'] = data
    return rep.dump()


def try_decode(data, cls):
    try:
        req = cls.load(data, strict=True)
    except ValueError as e:
        raise ASN1ParsingError(e)
    return req.pretty_name



# Copyright (C) 2013, Red Hat, Inc.
# All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import os

from setuptools import setup

install_requires = [
    'asn1crypto>=0.23',
    'dnspython'
]

extras_require = {
    "tests": ["pytest", "coverage", "WebTest"],
    "test_pep8": ['flake8', 'flake8-import-order', 'pep8-naming']
}


def read(fname):
    fname = os.path.join(os.path.dirname(__file__), fname)
    with open(fname) as f:
        return f.read()


# in chronological order
authors = {
    "Nalin Dahyabhai": "nalin@redhat.com",
    "Nathaniel McCallum": "npmccallum@redhat.com",
    "Christian Heimes": "cheimes@redhat.com",
    "Robbie Harwood": "rharwood@redhat.com",
}
setup(
    name="kdcproxy",
    version="0.4.2",
    author=", ".join(authors.keys()),
    author_email=", ".join(authors.values()),
    description=("A kerberos KDC HTTP proxy WSGI module."),
    license="MIT",
    keywords="krb5 proxy http https kerberos",
    url="http://github.com/npmccallum/kdcproxy",
    packages=['kdcproxy', 'kdcproxy.config'],
    long_description=read('README'),
    install_requires=install_requires,
    extras_require=extras_require,
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Web Environment",
        "Intended Audience :: System Administrators",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Topic :: Internet :: Proxy Servers",
    ],
)
